﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;

namespace FlexGridBenchmark.Utils
{
    public static class DataGenerator
    {
        private static readonly Type[] DataTypes = new Type[]
        {
            typeof(int),
            typeof(string),
            typeof(double),
            typeof(DateTime),
            typeof(decimal),
            typeof(bool),
            typeof(int) // for DataMap
        };

        private static readonly string Chars = " $%#@!*?;:^&abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890";

        static Random _random = null;

        public static Data Generate(int rowCount, int columnCount)
        {
            _random = new Random(0);

            Dictionary<int, string>[] dataMaps = GetRandomDataMaps(columnCount);

            DataTable dataTable = GetRandomDataTable(rowCount, columnCount, dataMaps);

            var data = new Data(dataTable, dataMaps, Array.LastIndexOf(DataTypes, typeof(int)), DataTypes.Length);

            return data;
        }

        private static DataTable GetRandomDataTable(int rowCount, int columnCount, Dictionary<int, string>[] dataMaps)
        {
            int dataTypesLength = DataTypes.Length;

            var dataTable = new DataTable();

            int firstDataMapColumnIndex = Array.LastIndexOf(DataTypes, typeof(int));

            for (int columnIndex = 0; columnIndex < columnCount; columnIndex++)
            {
                int loopIndex = columnIndex / dataTypesLength;
                int typeIndex = columnIndex % dataTypesLength;
                Type columnType = DataTypes[typeIndex];

                string columnName = string.Empty;
                if (typeIndex != firstDataMapColumnIndex)
                {
                    columnName = $"{columnType} {loopIndex}";
                }
                else
                {
                    columnName = $"DataMap {loopIndex}";
                }

                dataTable.Columns.Add(columnName, columnType);
            }

            for (int rowIndex = 0; rowIndex < rowCount; rowIndex++)
            {
                DataRow dataRow = dataTable.NewRow();
                for (int columnIndex = 0; columnIndex < columnCount; columnIndex++)
                {
                    DataColumn dataColumn = dataTable.Columns[columnIndex];

                    object value = null;
                    int typeIndex = columnIndex % dataTypesLength;
                    if (typeIndex != firstDataMapColumnIndex)
                    {
                        value = GetRandomValueByType(dataColumn.DataType);
                    }
                    else
                    {
                        int loopIndex = columnIndex / dataTypesLength;
                        value = GetRandomValueFromDataMap(dataMaps, loopIndex);
                    }
                    dataRow[dataColumn] = value;
                }

                dataTable.Rows.Add(dataRow);
            }

            return dataTable;
        }

        private static Dictionary<int, string>[] GetRandomDataMaps(int columnCount)
        {
            int dataTypesLength = DataTypes.Length;
            int dataMapCount = columnCount / dataTypesLength;

            var dataMaps = new Dictionary<int, string>[dataMapCount];

            for (int dataMapIndex = 0; dataMapIndex < dataMapCount; dataMapIndex++)
            {
                dataMaps[dataMapIndex] = GetRandomDataMap();
            }

            return dataMaps;
        }

        private static Dictionary<int, string> GetRandomDataMap()
        {
            int itemCount = _random.Next(100, 201);
            var dataMap = new Dictionary<int, string>();
            for (int i = 0; i < itemCount; i++)
            {
                int key = 0;
                int existingKeyCount = 0;
                int keyCount = dataMap.Count;
                do
                {
                    key = GetRandomInt32();
                }
                while (dataMap.ContainsKey(key) && existingKeyCount++ <= keyCount);

                if (existingKeyCount > keyCount)
                {
                    break;
                }

                dataMap.Add(key, GetRandomString());
            }

            return dataMap;
        }

        private static object GetRandomValueByType(Type dataType)
        {
            switch (Type.GetTypeCode(dataType))
            {
                case TypeCode.Int32:
                    {
                        return GetRandomInt32();
                    }
                case TypeCode.String:
                    {
                        return GetRandomString();
                    }
                case TypeCode.Double:
                    {
                        return GetRandomDouble();
                    }
                case TypeCode.Decimal:
                    {
                        return GetRandomDecimal();
                    }
                case TypeCode.DateTime:
                    {
                        return GetRandomDateTime();
                    }
                case TypeCode.Boolean:
                    {
                        return GetRandomBoolean();
                    }
            }

            return null;
        }

        private static int GetRandomInt32()
        {
            byte[] buffer = new byte[sizeof(int)];
            _random.NextBytes(buffer);
            return BitConverter.ToInt32(buffer, 0);
        }

        private static string GetRandomString()
        {
            int stringLength = _random.Next(5, 101);
            int charsLength = Chars.Length;
            char ch;
            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i < stringLength; i++)
            {
                int charIndex = _random.Next(0, charsLength);
                ch = Chars[charIndex];
                stringBuilder.Append(ch);
            }
            return stringBuilder.ToString();
        }

        private static double GetRandomDouble()
        {
            byte[] buffer = new byte[sizeof(double)];
            _random.NextBytes(buffer);
            return BitConverter.ToDouble(buffer, 0);
        }

        public static decimal GetRandomDecimal()
        {
            int a = (int)(uint.MaxValue * _random.NextDouble());
            int b = (int)(uint.MaxValue * _random.NextDouble());
            int c = (int)(uint.MaxValue * _random.NextDouble());
            bool n = _random.NextDouble() > 0.5;
            byte s = (byte)(29 * _random.NextDouble());
            return new decimal(a, b, c, n, s);
        }

        private static DateTime GetRandomDateTime()
        {
            int year = _random.Next(1900, 9999);
            int month = _random.Next(1, 12);
            int daysInMonth = DateTime.DaysInMonth(year, month);
            int day = _random.Next(1, daysInMonth + 1);
            DateTime dt = new DateTime(year, month, day).AddMilliseconds(GetRandomInt32() % 86400000);
            return dt;
        }

        private static bool GetRandomBoolean()
        {
            return _random.Next(0, 2) == 1 ? true : false;
        }

        private static int GetRandomValueFromDataMap(Dictionary<int, string>[] dataMaps, int dataMapIndex)
        {
            Dictionary<int, string> dataMap = dataMaps[dataMapIndex];
            int dataMapItemCount = dataMap.Count;

            int dataMapItemIndex = _random.Next(0, dataMapItemCount);

            return dataMap.ElementAt(dataMapItemIndex).Key;
        }
    }
}
